char *body_filter_rcs = "$Id: body_filter.c,v 2.0 2000/08/05 19:30:00 ossi Exp $";
/* (C) 2000 by Oswald Buddenhagen <junkex@ossi.cjb.net>.
 * Based on body_filter.c from LeanWeb, (C) 1999 by Carsten Clasohm. 
 * Distributed under the GNU General Public License; see the README file.
 * This code comes with NO WARRANTY. http://www.junkbusters.com/ht/en/gpl.html
 */

#define dbg if(0) printf

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "jcc.h"

#ifndef min
#define min(a, b)  (((a) < (b)) ? (a) : (b))
#endif

#ifndef max
#define max(a, b)  (((a) > (b)) ? (a) : (b))
#endif

extern int strcmpic(), strncmpic();
extern FILE *logfp;

/* find rule in buf, return match offset in moff, match length in mlen,
   return 0 if no match, 1 if partial match and 2 if complete match */
/* it's full of ugly gotos, but it's FAST */
/* this is case-sensitive!!! */
int matchen(char *buff, int blen, char *rule, char **moff, int *mlen)
{
	char r, b, *buf=buff-1, *rp;
	int bp, len=blen;

	for(;;) {
		bp=0; rp = rule; r = *rp;
		if(r=='\t') goto xmasp;
		if(r=='\n') goto xmanl;
		do {
			buf++;
			if(--len<0) goto ret0;
		} while(*buf!=r);
		do {
			rp++;
			if(++bp>len) goto ppartial;
			b = buf[bp];
next:			if(!(r=*rp)) goto complete;
			if(r=='\t') goto masp;
			if(r=='\n') goto manl;
normal:		} while(b==r);
		continue;
		
xmasp:		buf++;
		if(--len<0) goto ret0;
		b = *buf;
masp:		if(b!=' ' && b!='\t') continue;
		rp++;
		do {
			if(++bp>len) goto ppartial;
			b = buf[bp];
		} while(b==' ' || b=='\t');
		goto next;
	}
	
xmanl:	buf++;
	if(--len<0) goto ret0;
	b = *buf;
manl:	while(b==' ' || b=='\t') {
		if(++bp>len) goto partial;
		b = buf[bp];
	}
	if(b=='\r') {
		if(++bp>len) goto partial;
		b = buf[bp];
	}
	goto normal;

ppartial:
	if(!*rp) {
complete:	*moff = buf;
		*mlen = bp;
		return 2;
	}
partial:
	*moff = buf;
	*mlen = 0;
	return 1;

ret0:	*moff = buf;
	*mlen = 0;
	return 0;
}

int init_bfilter(struct client_state *csp, struct http_request *http, struct bfState *sts)
{
	char *domain = http->host;
	char *path = http->path;
	int nrul = 0, lrd, ld = strlen(domain);
	struct bfilter_spec *fi;
	struct srule *ru, **rptr = &(sts->rules);
	struct smatch *ma, **mptr = &(sts->matches);

dbg("init_bfilter for http://%s%s\n",domain,path);

	for(fi=(struct bfilter_spec *)csp->bflist->f; fi; fi=fi->next) {
		lrd = strlen(fi->domain);
		if(lrd <= ld && !strcmpic(domain + ld - lrd, fi->domain) &&
		  (!fi->path || !strncmpic(path, fi->path, strlen(fi->path)))) {
			if(fi->start[0]) {
				ru = malloc(sizeof(*ru));
				if(!ru) break;
				*rptr = ru;
				rptr = &(ru->next);
				ru->spec = fi;
				ru->soff = 0;
				ru->next = 0;
			} else {
				ma = malloc(sizeof(*ma));
				if(!ma) break;
				*mptr = ma;
				mptr = &(ma->next);
				ma->spec = fi;
				ma->moff = 0;
				ma->soff = 0;
				ma->coff = 0;
				ma->next = 0;
dbg("unconditional start match\n");
			}
			nrul++;
		}
	}
	return nrul;
}

void cleanup_bfilter(struct bfState *sts)
{
	struct srule *rptr, *trptr;
	struct smatch *mptr, *tmptr;

	while(sts->matches) {
		tmptr = sts->matches->next;
		free(sts->matches);
		sts->matches = tmptr;
	}
	while(sts->rules) {
		trptr = sts->rules->next;
		free(sts->rules);
		sts->rules = trptr;
	}
}

int filter_body(struct bfState *sts, int blen,
	char *buf, int len, char **rbuf, int *rlen)
{
	int foff, nfoff, deleted, tp, off, ma, mlen, imoff, ecoff, sl, ldiff;
	char *tbuf, *tbufp, *moff;
	struct bfilter_spec *fi;
	struct srule *rptr;
	struct smatch *mptr, **mpptr, *tmptr;
	size_t bdiff;

dbg("bodyfilter for ===\n%.*s\n===\n",len,buf);

	/* create lowercase copy */
#ifdef HAVE_ALLOCA
	tbuf=alloca(len);
#else
	tbuf=malloc(len);
#endif
	if(!tbuf) return len;
	bdiff = tbuf - buf;
	for(tp=0; tp<len; tp++)
		tbuf[tp] = tolower(buf[tp]);

	/* find all occurances of the starting strings of all rules */
	for(rptr = sts->rules; rptr; rptr = rptr->next) {
		fi = rptr->spec;
		mpptr = &(sts->matches);
		for(;;) {
			off = rptr->soff - sts->fileoff;
			ma = matchen(buf + bdiff + off, len - off, fi->start, &moff, &mlen);
dbg("start match ===\n%s\n===\n => %d \"%.*s\"\n===\n",fi->start,ma,mlen,moff-bdiff);
			imoff = moff - (buf + bdiff) + sts->fileoff;
			rptr->soff = imoff + mlen;
			if((ma==0) || ((ma==1) && (blen<BF_MIN_BLEN))) break;
			while(*mpptr) {
				if(((*mpptr)->moff == imoff) && ((*mpptr)->spec == fi)) {
					mptr = *mpptr;
					goto setten;
				}
				if((*mpptr)->moff > imoff) break;
				mpptr = &((*mpptr)->next);
			}
			mptr = malloc(sizeof(*mptr));
			if(!mptr) break;
			mptr->next = *mpptr; *mpptr = mptr;
			mptr->spec = fi;
			mptr->moff = imoff;
setten:			if(ma==2) {
				mptr->soff = rptr->soff;
				mptr->coff = fi->deleteStart? imoff : rptr->soff;
				continue;
			}
			mptr->soff = sts->fileoff + len;
			mptr->coff = -1;
			break;
		}
	}

	deleted = 0;
	foff = sts->fileoff;

	for(; mptr = sts->matches; tmptr = sts->matches->next, 
				free(sts->matches), sts->matches = tmptr) {
dbg("found start match\n");
		if(mptr->moff < foff) continue;	/* obsoleted */
		if(mptr->coff == -1) {
			if(mptr->soff < sts->fileoff+len) continue; /* obsoleted */
dbg(" postponing partial start match\n");
			break;	/* wait for next block */
		}
dbg(" is up-to-date\n");
		fi = mptr->spec;
		if(fi->end) {
			off = mptr->soff - sts->fileoff;
			if(fi->end[0]) {
				ma = matchen(buf + bdiff + off, len - off, fi->end, &moff, &mlen);
dbg(" end match ===\n%s\n===\n => %d \"%.*s\"\n===\n",fi->end,ma,mlen,moff-bdiff);
				if(ma!=2) {
					if(blen < BF_MIN_BLEN) continue; /* not matched */
dbg("  postponing end match\n");
					break;	/* wait for next block */
				}
				imoff = moff - (buf + bdiff) + sts->fileoff;
				ecoff = imoff + (fi->deleteEnd? mlen : 0);
dbg("  end match parameters: imoff=%d mlen=%d ecoff=%d\n", imoff, mlen,ecoff);
			} else {
				if(blen) break;	/* wait for last block */
				ecoff = imoff = sts->fileoff + len;
dbg(" unconditional end match\n");
			}

			if(fi->cond) {
				ma = matchen(buf + bdiff + off, imoff - mptr->soff, fi->cond, &moff, &mlen);
				if(ma!=2) continue; /* not matched */
dbg(" cut-condition match\n");
			}
		} else
			ecoff = mptr->soff;

		if(fi->subst) {
			sl = strlen(fi->subst);
			if(sl <= (ecoff - mptr->coff)) {
				memcpy(buf + mptr->coff - sts->fileoff, fi->subst, sl);
				mptr->coff += sl;
			}
		}

dbg(" match parameters: moff=%d coff=%d soff=%d ecoff=%d\n",mptr->moff,mptr->coff,mptr->soff,ecoff);

		if(deleted) {
			tbufp = buf + foff - sts->fileoff;
			memcpy(tbufp - deleted, tbufp, mptr->coff - foff);
		}
dbg(" deleting \"%.*s\"  ===\n", ecoff - mptr->coff, buf + mptr->coff - sts->fileoff);
		deleted += (ecoff - mptr->coff);
		foff = ecoff;
	}

#ifndef HAVE_ALLOCA
	free(tbuf);
#endif

	nfoff = sts->matches? sts->matches->moff : sts->fileoff + len;
	if(deleted) {
		tbufp = buf + foff - sts->fileoff;
		memcpy(tbufp - deleted, tbufp, nfoff - foff);
	}
	foff = sts->fileoff; sts->fileoff = nfoff;
	ldiff = (nfoff - foff);
	*rlen = len - ldiff; *rbuf = buf + ldiff;
dbg("returning === \"%.*s\" ===\n", ldiff - deleted, buf);
	return ldiff - deleted;
}
